#!/usr/bin/env python

import os
import pipes
import shlex
import tarfile
import tempfile

from ..action import Action
from ..process import Process

class SSHTarget:
	def __init__ (self, host, port, user, directory, options):
		extra = shlex.split (options.get ('extra', ''))
		remote = str ((user or os.getusername ()) + '@' + (host or 'localhost'))

		self.directory = directory
		self.tunnel = ['ssh', '-T', '-p', str (port or 22)] + extra + [remote]

	def read (self, logger, relative):
		command = 'test -d \'{0}\' && ( test ! -f \'{1}\' || cat \'{1}\' )'.format (pipes.quote (self.directory), pipes.quote (self.directory + '/' + relative))
		result = Process (self.tunnel + [command]).execute ()

		if not result:
			logger.debug (result.err)

			return None

		return result.out

	def send (self, logger, work, actions):
		with tempfile.TemporaryFile () as archive:
			to_add = False
			to_del = []

			# Append files to temporary TAR archive or deletion list
			with tarfile.open (fileobj = archive, mode = 'w') as tar:
				for action in actions:
					if action.type == Action.ADD:
						tar.add (os.path.join (work, action.path), action.path)

						to_add = True
					elif action.type == Action.DEL:
						to_del.append (self.directory + '/' + action.path)

			archive.seek (0)

			# Send and delete files on remote host
			if to_add:
				result = Process (self.tunnel + ['tar xC \'' + pipes.quote (self.directory) + '\'']).set_input (archive.read ()).execute ()

				if not result:
					logger.warning ('Couldn\'t push files to SSH target.')
					logger.debug (result.err)

					return False

			if len (to_del) > 0:
				result = Process (self.tunnel + ['sh']).set_input (';'.join (['rm -f \'' + pipes.quote (path) + '\'' for path in to_del])).execute ()
				
				if not result:
					logger.warning ('Couldn\'t delete files from SSH target.')
					logger.debug (result.err)

					return False

		return True



#!/usr/bin/env python

import argparse
import logging
import os
import sys

sys.path.append (os.path.dirname (__file__))

from src import Logger, deploy

def main ():
	# Parse command line options
	parser = argparse.ArgumentParser (description = 'Perform incremental deployment from Git/plain workspace to FTP/SSH/local folder.')
	parser.add_argument ('name', nargs = '*', help = 'Deploy to specified named location (* = everywhere)')
	parser.add_argument ('-a', '--append', action = 'append', default = [], help = 'Manually append file or directory to locations', metavar = 'PATH')
	parser.add_argument ('-b', '--base', default = '.', help = 'Change base directory', metavar = 'DIR')
	parser.add_argument ('--extra-append', action = 'append', default = [], help = argparse.SUPPRESS)
	parser.add_argument ('-e', '--environment', default = '.creep.env', help = 'Use specified environment file', metavar = 'FILE')
	parser.add_argument ('-f', '--rev-from', help = 'Initial version used to compute diff', metavar = 'REV')
	parser.add_argument ('-d', '--definition', default = '.creep.def', help = 'Use specified definition file', metavar = 'FILE')
	parser.add_argument ('-q', '--quiet', dest = 'level', action = 'store_const', const = logging.CRITICAL + 1, default = logging.INFO, help = 'Quiet mode, don\'t display anything but errors')
	parser.add_argument ('-r', '--remove', action = 'append', default = [], help = 'Manually remove file or directory from locations', metavar = 'PATH')
	parser.add_argument ('--extra-remove', action = 'append', default = [], help = argparse.SUPPRESS)
	parser.add_argument ('-t', '--rev-to', help = 'Target version used to compute diff', metavar = 'REV')
	parser.add_argument ('-v', '--verbose', dest = 'level', action = 'store_const', const = logging.DEBUG, default = logging.INFO, help = 'Verbose mode, display extra information')
	parser.add_argument ('-y', '--yes', action = 'store_true', help = 'Always answer yes to prompts')

	args = parser.parse_args ()

	if not deploy.execute (
		Logger.build (args.level),
		args.base,
		args.definition,
		args.environment,
		args.name,
		args.append + args.extra_append,
		args.remove + args.extra_append,
		args.rev_from,
		args.rev_to,
		args.yes):
		return 1

	return 0

if __name__ == '__main__':
	sys.exit (main ())



#!/usr/bin/env python

import os
import shutil

from .. import path

class FileTarget:
	def __init__ (self, directory):
		self.directory = directory

	def read (self, logger, relative):
		if not os.path.isdir (self.directory):
			return None

		source = os.path.join (self.directory, relative)

		if not os.path.isfile (source):
			return ''

		if not os.access (source, os.R_OK):
			return None

		with open (source, 'rb') as file:
			return file.read ()

	def send (self, logger, work, actions):
		for action in actions:
			if action.type == action.ADD:
				if not path.duplicate (os.path.join (work, action.path), self.directory, action.path):
					logger.warning ('Can\'t copy file "{1}" to target directory "{0}"'.format (self.directory, action.path))

			elif action.type == action.DEL:
				if not path.remove (self.directory, action.path):
					logger.warning ('Can\'t remove file "{1}" from target directory "{0}"'.format (self.directory, action.path))

		return True



#!/usr/bin/env python

import logging
import re

class ColorStreamHandler (logging.StreamHandler):
	COLOR_BEGIN = '(('
	COLOR_END = '))'
	RESET = '\033[0m'

	def __init__ (self, *args, **kwargs):
		super (ColorStreamHandler, self).__init__ (*args, **kwargs)

		self.colors = {
			'black':	'\033[0;30m',
			'maroon':	'\033[0;31m',
			'green':	'\033[0;32m',
			'olive':	'\033[0;33m',
			'navy':		'\033[0;34m',
			'purple':	'\033[0;35m',
			'teal':		'\033[0;36m',
			'silver':	'\033[0;37m',
			'grey':		'\033[1;30m',
			'red':		'\033[1;31m',
			'lime':		'\033[1;32m',
			'yellow':	'\033[1;33m',
			'blue':		'\033[1;34m',
			'fuchsia':	'\033[1;35m',
			'cyan':		'\033[1;36m',
			'white':	'\033[1;37m',
			'reset':	self.RESET
		}

		self.end = getattr (self, 'terminator', '\n')
		self.levels = {logging.CRITICAL: 'red', logging.DEBUG: 'olive', logging.ERROR: 'red', logging.INFO: 'white', logging.WARNING: 'yellow'}
		self.tags = re.compile (re.escape (self.COLOR_BEGIN) + '(' + '|'.join (self.colors.keys ()) + ')' + re.escape (self.COLOR_END))

	def is_tty (self):
		isatty = getattr (self.stream, 'isatty', None)

		return isatty and isatty ()

	def emit (self, record):
		try:
			message = self.format (record)
			stream = self.stream

			if self.is_tty ():
				message = self.COLOR_BEGIN + self.levels.get (record.levelno, 'reset') + self.COLOR_END + message + self.RESET
				message = self.tags.sub (lambda m: self.colors.get (m.group (1), self.RESET), message)
			else:
				message = self.tags.sub ('', message)

			stream.write (message)
			stream.write (self.end)

			self.flush ()

		except (KeyboardInterrupt, SystemExit):
			raise
		except:
			self.handleError (record)

class IndentLoggerAdapter (logging.LoggerAdapter):
	def __init__ (self, logger, extra):
		super (IndentLoggerAdapter, self).__init__ (logger, extra)

		self.indent = 0

	def enter (self):
		self.indent = min (self.indent + 1, 8)

	def leave (self):
		self.indent = max (self.indent - 1, 0)

	def process (self, msg, kwargs):
		return '| ' * self.indent + msg, kwargs

class Logger:
	@staticmethod
	def build (level):
		formatter = logging.Formatter ('%(levelname)s: %(message)s')

		console = ColorStreamHandler ()
		console.setFormatter (formatter)

		logger = logging.getLogger ()
		logger.addHandler (console)
		logger.setLevel (level)

		return IndentLoggerAdapter (logger, {})



#!/usr/bin/env python

import ftplib
import io
import itertools
import os

from ..action import Action
from .. import path

class FTPTarget:
	def __init__ (self, host, port, user, password, directory, options):
		self.directory = directory
		self.host = host or 'localhost'
		self.options = options
		self.port = port or 21
		self.password = password
		self.user = user

	def connect (self, logger):
		ftp = ftplib.FTP ()
		ftp.connect (self.host, self.port)

		try:
			if self.user is not None:
				ftp.login (self.user, self.password)

			if self.directory:
				ftp.cwd (self.directory)

			ftp.set_pasv (self.options.get ('passive', True))
		except ftplib.all_errors as e:
			if e.message.startswith ('530 '):
				logger.debug ('Can\'t authenticate as \'{0}\' on remote FTP: \'{1}\''.format (self.user, e.message))
			elif e.message.startswith ('550 '):
				logger.debug ('Can\'t access folder \'{0}\' on remote FTP: \'{1}\''.format (self.directory, e.message))
			else:
				logger.debug ('Unknown FTP error: \'{0}\''.format (e.message))

			ftp.quit ()

			return None

		return ftp

	def escape (self, path):
		return path # FIXME: wrong escape [ftp-escape]

	def read (self, logger, relative):
		ftp = self.connect (logger)

		if ftp is None:
			return None

		with io.BytesIO () as buffer:
			try:
				ftp.retrbinary ('RETR ' + self.escape (relative), buffer.write)

				return buffer.getvalue ()

			except ftplib.all_errors as e:
				if e.message.startswith ('550 '): # no such file or directory
					return ''

				logger.debug ('Can\'t read file from FTP remote: {0}'.format (e.message))

				return None

			finally:
				ftp.quit ()

	def send (self, logger, work, actions):
		ftp = self.connect (logger)

		if ftp is None:
			return None

		try:
			# Group actions by parent path
			commands = [(head, tail, type) for ((head, tail), type) in ((os.path.split (action.path), action.type) for action in actions)]

			for (directory, files) in itertools.groupby (commands, lambda command: command[0]):
				# Must create directory before uploading files to it
				create = True
				names = path.explode (directory)

				# Append or delete files
				for (head, tail, type) in files:
					target = self.escape (head and head + '/' + tail or tail)

					if type == Action.ADD:
						# Create missing parent directories
						if create:
							for parent in ('/'.join (names[0:n + 1]) for n in range (0, len (names))):
								try:
									ftp.mkd (parent)
								except ftplib.all_errors as e:
									if not e.message.startswith ('550 '):
										raise e

							create = False

						# Upload current file
						with open (os.path.join (work, head, tail), 'rb') as file:
							ftp.storbinary ('STOR ' + target, file)

					elif type == Action.DEL:
						# Delete file if exists
						try:
							ftp.delete (target)
						except ftplib.all_errors as e:
							if not e.message.startswith ('550 '):
								raise e

		except ftplib.all_errors as e:
			logger.debug ('Can\'t deploy to FTP remote: {0}'.format (e.message))

			return False

		finally:
			ftp.quit ()

		return True



#!/usr/bin/env python

import os
import shutil

"""
Copy file and create parent directories if needed.
source:	path to source file
base:	base directory of target file (won't be created)
target:	path to target file relative to base directory (will be created)
"""
def duplicate (source, base, target):
	if not os.path.isdir (base):
		return False

	destination = os.path.join (base, target)
	directory = os.path.dirname (destination)

	if not os.path.isdir (directory):
		os.makedirs (directory)

	shutil.copy (source, destination)

	return True

"""
Explode path into separate components.
path:	input path
return:	path components
"""
def explode (path):
	names = []
	tail = '.'

	while tail != '':
		(path, tail) = os.path.split (path)

		if tail != '':
			names.insert (0, tail)
		elif path != '':
			names.insert (0, path)

	return names

"""
Remove file from given directory if exists.
base:	base directory of target file
target:	path to target file relative to base directory
"""
def remove (base, target):
	remove = os.path.join (base, target)

	if not os.path.isfile (remove):
		return False

	os.remove (remove)

	return True



#!/usr/bin/env python

class Action:
	ADD = 1
	DEL = 2
	ERR = 3
	NOP = 4

	def __init__ (self, path, type):
		self.path = path
		self.type = type

	def order (self):
		if self.type == Action.DEL:
			return 0
		elif self.type == Action.ADD:
			return 1
		elif self.type == Action.NOP:
			return 2
		else:
			return 3



#!/usr/bin/env python

from codecs import open
from os import path
from setuptools import find_packages, setup

import creep

with open (path.join (path.abspath (path.dirname (__file__)), 'README.md'), encoding = 'utf-8') as file:
	long_description = file.read ()

setup(
	name = 'creep',
	version = creep.__version__,
	description = 'Incremental FTP/SSH deployment tool',
	long_description = long_description,
	url = 'https://github.com/r3c/creep',
	author = 'Remi Caput',
	author_email = 'python.org+creep@mirari.fr',
	license = 'MIT',
	keywords = 'deploy deployment incremetal ftp ssh git',
	classifiers = [
		'Development Status :: 4 - Beta',
		'Environment :: Console',
		'Intended Audience :: Developers',
		'Intended Audience :: System Administrators',
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 2.7',
		'Topic :: Software Development :: Build Tools',
		'Topic :: System :: Archiving :: Mirroring',
		'Topic :: System :: Software Distribution'
	],
	install_requires = ['setuptools>=1.0'],
	packages = [
		'creep',
		'creep.src',
		'creep.src.sources',
		'creep.src.targets'
	],
	entry_points = {
		'console_scripts':	['creep = creep.creep:main']
	}
)
